/** 
    @file
    @ingroup calib
    @brief   calibration routines

    This header contains some BPM calibration routines
*/

/** @addtogroup calib */
/** @{ */

#ifndef BPMCALIBRATION_H__
#define BPMCALIBRATION_H__

/* -----------------------------------------------------------------------------
// includes
// -------------------------------------------------------------------------- */
#include <math.h>
#include <bpm/bpm_defs.h>
#include <bpm/bpm_interface.h>

/* -----------------------------------------------------------------------------
// macro definitions
// -------------------------------------------------------------------------- */

/* -----------------------------------------------------------------------------
// typedefs, enums and other declarations
// -------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
// function prototypes and declarations
// -------------------------------------------------------------------------- */

  /**
     This routine basically defines the calibration steps and returns them into
     the array of beam structures. It needs an array of processed waveform structures,
     of dimension npulses from a single BPM. From this it determines the corresponding 
     corrector/mover steps and puts them back into the array of beam structures given 
     the bpm configurations.  

     Startpulse and stoppulse have to be in the first and last calib steps &
     will need some extensive error checking for e.g. missed calibration steps...

     NOTE: This is not definitive yet - more checking, etc. required! 
            - DDC or FIT?
            - Sign errors? 
            - not robust to missing steps

     @param proc       array of processed waveforms for a single bpm, so array of pulses
     @param cnf        array of bpm configuration structures
     @param npulses    number of pulses in the calibration
     @param startpulse start of calibration range
     @param stoppulse  stop of calibration range
     @param angle
     @param startpos   start position of calibration
     @param endpos     end position of calibration
     @param num_steps  number of calibration steps
     @param beam       the returned beamconf array which represents where the beam is 
                       supposed to be in each bpm during each calibration step

     @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */
  EXTERN int setup_calibration( bpmconf_t *cnf, bpmproc_t *proc, int npulses, int startpulse, 
				int stoppulse, double angle, double startpos, double endpos, 
				int num_steps, beamconf_t *beam );

  
  /**
     Gets the calibration constants from an array of npulses of beam positions
     and processed waveform structures and returns an updated calibration structure.
     Note that this routine updates the IQ phase, the position scale and the tilt scale 
     but DOES NOT touch the frequency, decay time or the t0Offset.

     @param  bpm     Bpm structures
     @param  beam    An array of beam structures, one for each pulse, so essentially 
                     this corresponds to where we expect the beam to be in each pulse,
                     so representing corrector positions or mover positions. This information 
                     should be filled by the routine setup_calibration( ... )
     @param  proc    An array of processed waveforms, one for each pulse, which correspond
                     to calculated positions that were calculated using IQ phase = 0 and
                     scales equal to 1.
     @param  npulses The number of pulses in the arrays 
     @param *cal     The returned calibration structure for the BPM that was calibrated 
     @return BPM_SUCCESS upon succes, BPM_FAILURE upon failure 
  */ 
  EXTERN int calibrate( bpmconf_t *bpm, beamconf_t *beam, bpmproc_t *proc, int npulses, 
			bpmcalib_t *cal );


  /**
     Gets the list of processed pulses and refills the calib structure with updated
     frequencies and decay constants
     
     NOT IMPLEMENTED YET !

     @param proc    array of processed waveforms
     @param npulses the number of pulses
     @param cal     the refilled calibration structure
     @return BPM_SUCCESS upon succes, BPM_FAILURE upon failure
 */
  EXTERN int update_freq_tdecay( bpmproc_t *proc, int npulses, bpmcalib_t *cal );


  
  /**
     The 2D arrays in this routine represent a set of collected pulses for all the
     bpms, so having beam[iBPM][iPulse], cnf[iBPM][iPulse] and proc[iBPM][iPulse],
     @param  npulses The number of pulses collected for calibration

     Used for mover calibrations with at least 2 spectator bpms.
     eats something of the sort bpms[bpmidx][pulseidx], for a number of pulse and. 
     nbpms specifies the total number of bpms involved in the regression
     bpmidx specifies the indexes of the bpms involve in the regression,
     bpmidx[0] gives the central bpm, for which the calibration is calculated
     the rest ( bpmidx[1] -> bpmidx[nbpms-1]) gives the indexes of the spectator
     bpms.

     NOT IMPLEMENTED YET !

     @param  beam
     @param  bpm
     @param  proc
     @param  npulses
     @param  nbpms   The total number of BPMs that will be used in the regression 
     @param  bpmidx  An array of bpm indexes, where bpmidx[0] corresponds to the
                     index of the bpm in the main array that we will calibrate, and
                     the rest of the indices corresponds to the BPMs we will
                     regress against, so basically the spectator BPMs,
                     when doing a mover calibration
     @param  cal     This structure is filled with the calculated iqphase,
                     and position and resolution scales
     @return BPM_SUCCESS upon success, BPM_FAILURE upon failure                              
  */

  EXTERN int calibrate_svd( beamconf_t **beam, bpmconf_t **bpm, bpmproc_t **proc, int npulses,
			    int nbpms, int *bpmidx, bpmcalib_t *cal );
  
  /**
     Save the given calibrations with the given filename.

     @param fname    The filename to save as
     @param bpm      BPM configs - to provide a name and index
     @param cal      The calibrations to save
     @param num_bpms The number of bpm cals to save
     @return BPM_SUCCESS upon succes, BPM_FAILURE upon failure
  */
  EXTERN int save_calibration( char *fname, bpmconf_t *bpm, bpmcalib_t *cal, int num_bpms );


  /**
     Load the calibration from the given filename.

     @param fname    The filename to load#
     @param bpm      BPM configs - to provide a name and index
     @param cal      The calibrations to load
     @param num_bpms The number of bpm cals to load

     @return BPM_SUCCESS upon succes, BPM_FAILURE upon failure
  */
  EXTERN int load_calibration( char *fname, bpmconf_t *bpm, bpmcalib_t *cal, int num_bpms );

#ifdef __cplusplus
}
#endif

#endif /* #ifndef BPMCALIBRATION_H__ */
/*@}*/
/* ================================ end of file ============================= */
