/** 
    @file     
    @ingroup dsp
    @brief   libbpm digital signal processing routines
    
    This header contains the definitions for the digital signal processing routines
    for libbpm. 
*/

/** @addtogroup dsp */
/** @{ */

#ifndef BPMDSP_H__
#define BPMDSP_H__

/* -----------------------------------------------------------------------------
// includes
// -------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "bpm/bpm_defs.h"
#include "bpm/bpm_messages.h"
#include "bpm/bpm_alloc.h"
#include "bpm/bpm_nr.h"

/* -----------------------------------------------------------------------------
// macro definitions
// -------------------------------------------------------------------------- */
#define BESSEL                 0x00001 /**< Bitmask for Bessel filter */
#define BUTTERWORTH            0x00002 /**< Bitmask for Butterworth filter */
#define CHEBYSHEV              0x00004 /**< Bitmask for Chebyshev filter */
#define RAISEDCOSINE           0x00008 /**< Bitmask for Raised Cosine filter */
#define RESONATOR              0x00010 /**< Bitmask for Resonator filter */
#define GAUSSIAN               0x00020 /**< Bitmask for Gaussian filter */

#define BILINEAR_Z_TRANSFORM   0x00100 /**< Get z poles via bilinear z transform from s plane */
#define MATCHED_Z_TRANSFORM    0x00200 /**< Get z poles via matches z transform from s plane */  
#define NO_PREWARP             0x00400 /**< Don't do the prewarp correction */

#define LOWPASS                0x01000 /**< Normalise filter as lowpass */
#define HIGHPASS               0x02000 /**< Normalise filter as highpass */
#define BANDPASS               0x04000 /**< Normalise filter as bandpass */
#define BANDSTOP               0x08000 /**< Normalise filter as bandstop */
#define NOTCH                  0x08000 /**< Normalise filter as notch filter (=bandstop) */
#define ALLPASS                0x10000 /**< Normalise filter as allpass ( resonator ) */

#define MAXORDER                   10  /**< Maximum filter order */
#define MAXPZ              2*MAXORDER  /**< Maximum number of poles and zeros */
#define FILT_EPS              1.0e-10  /**< A small number used in bpmdsp */
#define MAX_RESONATOR_ITER         50  /**< Maximum iterations in resonator poles calculation */

/* -----------------------------------------------------------------------------
// typedefs, enums and other declarations
// -------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif

/**
 * The filter representation in the complex plane (poles/zeros).
 */
typedef struct {
  int        npoles;       /**< The number of filter poles */
  int        nzeros;       /**< The number of filter zeros */
  complex_t  pole[MAXPZ];  /**< Array of the filter's complex poles */
  complex_t  zero[MAXPZ];  /**< Array of the filter's complex zeros */
} filterrep_t;

/**
 * The filter structure.
 */
typedef struct {
  char          name[80];      /**< The filter's name */

  unsigned int  options;       /**< type and option bits for filter */
  int           order;         /**< filter order                    */
  
  double        fs;            /**< sampling frequency */
  double        f1;            /**< first frequency ( left edge for bandpass/stop ) */
  double        f2;            /**< right edge for bandpass/stop ( undef for low/highpass ) */

  double        alpha1;        /**< rescaled f1 */
  double        alpha2;        /**< rescaled f2 */

  double        w_alpha1;      /**< warped alpha1 */
  double        w_alpha2;      /**< warped alpha2 */
  
  double        cheb_ripple;   /**< ripple for chebyshev filters */
  double        Q;             /**< Q factor for resonators */

  complex_t     dc_gain;       /**< Complex DC gain of the filter               */
  complex_t     fc_gain;       /**< Complex Center frequency gain of filter     */
  complex_t     hf_gain;       /**< Complex High frequency (fNy) gain of filter */
  double        gain;          /**< Actual Filter gain */
  
  filterrep_t  *cplane;        /**< pointer to complex filter representation, poles and zeros */

  int           IsFIR;         /**< filter is FIR */
  int           nxc;           /**< number of x coefficients */
  double        xc[MAXPZ+1];   /**< pointer to array of x coefficients */
  int           nyc;           /**< number of y coefficients (for IIR filters) */
  double        yc[MAXPZ+1];   /**< pointer to array of y coefficients */

  double        xv[MAXPZ+1];   /**< filter x buffer, used in apply_filter */
  double        yv[MAXPZ+1];   /**< filter y buffer, used in apply_filter */

  int           ns;            /**< number of samples of waveforms to be filtered */
  double       *wfbuffer;      /**< waveform buffer for filter computations, allocated once ! */
  
} filter_t;

  
/* -----------------------------------------------------------------------------
// function prototypes and declarations
// -------------------------------------------------------------------------- */

  /**
     Creates the filter.
     @param name    a name for the filter
     @param options filter specification and options bitword 
     @param order   filter order
     @param ns      number of samples of the waveforms
     @param fs      sampling frequency
     @param f1      first frequency
     @param f2      optional second frequency ( bandpass/bandstop )
     @param par     optional parameter
                    - for chebyshev : ripple in dB
                    - for resonator : Q factor
     @return A pointer to the created filter structure, memory is allocated 
             on the heap inside this routine, the user has to take of deleting
             it using delete_filter().
  */
  EXTERN filter_t*    create_filter( char name[], unsigned int options, int order,
				     int ns, double fs, double f1, double f2, double par );

  /**
     Apply the filter to the given waveform.
     Note that the filter is applied in place, the user has to make a copy
     of the waveform if he/she wants to keep the original before applying the
     filter. The number of samples in the waveform has to be set in advance
     when creating the filter, it is stored in the filter structure (f->ns).

     @param  f   pointer to a filter that was created using create_filter
     @param  wf  an array containing the waveform to be filtered
     @return BPM_SUCCESS upon success and BPM_FAILURE upon failure
  */
  EXTERN int          apply_filter( filter_t *f, double *wf );


  /**
     Prints the filter to the given file pointer.
     @param of the filepointer, use "stdout" to print to the terminal
     @param f  the filter to be printed
     @return void
  */  
  EXTERN void         print_filter( FILE *of, filter_t *f );

  /**
     Clears the memory that was allocated on the heap for the filter f.
     @param f a pointer to the filter
     @return void
   */
  EXTERN void         delete_filter( filter_t *f );


  /**
     This routine fills the given wf with the step response of the filter. The step 
     response is defined as wf[i] = 0. for i < itrig and wf[i] = 1. for i >= itrig.

     @param f     a pointer to the filter to use
     @param wf    pointer to a waveform which will be overwritten with the step response
     @param itrig the sample number in the waveform which will have the step
     @return BPM_SUCCESS upon succes and BPM_FAILURE upon failure
  */
  EXTERN int filter_step_response( filter_t *f, double *wf, int itrig );

  /**
     This routine fills the given wf with the impulse response of the filter. The 
     impulse response is defined as wf[i] = 1. for i == itrig and wf[i] = 0. elsewhere.

     @param f     a pointer to the filter to use
     @param wf    pointer to a waveform which will be overwritten with the impulse response
     @param itrig the sample number in the waveform which will have the impulse
     @return BPM_SUCCESS upon succes and BPM_FAILURE upon failure
  */
  EXTERN int filter_impulse_response( filter_t *f, double *wf, int itrig );
  


  /**
     This routine returns a pointer to a filter representation filterrep_t in the 
     s plane for Butterworth, Chebyshev and Bessel filters. It need an initialised
     filter structure which has the filter type and the order set. Memory
     is allocated for this routine on the heap, so the user is responsible to 
     delete this memory using free(). 

     @param f the initialised filter with the correct options in f->options
     @return the filter representation in the s plane
  */
  EXTERN filterrep_t* create_splane_representation( filter_t *f );


  /**
     This routine returns a pointer to a filter representation filterrep_t in the 
     z plane for resonance filters. It needs an initialised filter structure which has 
     the filter type and the Q factor set. Memory is allocated for this routine on the
     heap, so the user is responsible to delete this memory using free(). 

     @param f the initialised filter with the correct options in f->options
     @return the filter representation in the z plane
  */
  EXTERN filterrep_t* create_resonator_representation( filter_t *f );

  /**
     This routine transforms the poles and zeros for Bessel, Chebyshev and Butterworth
     filters to the z plane either via matched z transform or bilinear z transform.
     This is set in f->options. Memory is allocated for this routine on the
     heap, so the user is responsible to delete this memory using free().

     @param f the filter, needs the options from it to check how to transform
     @param s filter s plane poles and zeros
     @return a pointer to the z plane representation
  */
  EXTERN filterrep_t* zplane_transform( filter_t *f, filterrep_t *s );


  /**
     Prints the filter representation in terms of poles and zeros to the filepointer.
     @param of the filepointer, use "stdout" to print to the terminal
     @param r  the filter representation to be printed
     @return void
   */
  EXTERN void print_filter_representation( FILE* of, filterrep_t *r );

  
  /**
     Normalises the Butterworth, Chebyshev or Bessel filters to be Bandpass/stop or Low/Highpass
     @param f  the filter
     @param s  the filter's representation in the s plane
     @return BPM_SUCCESS upon success or BPM_FAILURE upon failure.
   */
  EXTERN int normalise_filter( filter_t *f, filterrep_t *s );


  /**
     Calculates the filter coefficients from the z plane representation.
     Before this routine is called, one has to make sure that the member cplane, which
     holds a pointer to the filter's representation in the complex plane is set.
     This routine than calculates the filter coefficients and stores them in 
     f->xc ( coefficients of x[n], x[n-1], x[n-2]...) and f->yc ( coefficients of 
     y[n-1], y[n-2], y[n-3], ... in case of IIR filters ). 

     @param f the filter, having it's f->cplane member set to the z plan representation
     @return BPM_SUCCESS upon success or BPM_FAILURE upon failure.
  */
  EXTERN int calculate_filter_coefficients( filter_t *f );
  

  /**
     Helper routine to expand a complex polynomial from a set of zeros.
     @param w array of complex zeros for the polynomial
     @param n nunber of zeros
     @param a array of coeffiecients for the polynomial that is returned
     @return BPM_SUCCESS upon success or BPM_FAILURE upon failure.
  */
  EXTERN int       _expand_complex_polynomial( complex_t *w, int n, complex_t *a );

  /**
     Helper routine to evaluate a complex polynomial for value z
     @param a array of coeffiecients for the polynomial that is returned
     @param n number of zeros
     @param z the value for which to evalute the polynomial
     @return the value of the polynomial for z ( complex_t )
  */
  EXTERN complex_t _eval_complex_polynomial( complex_t *a, int n, complex_t z );

#ifdef __cplusplus
}
#endif

#endif /* #ifndef __BPM_FILTER_H_LOADED */
/*@}*/
/* ================================ end of file ============================= */

