/** 
   @file      
   @ingroup orbit
   @brief   libbpm orbit generation routines

   This header contains beam oribit generation routines, so this 
   includes also calibration scans etc...
*/

/** @addtogroup orbit */
/** @{ */

#ifndef BPMORBIT_H__
#define BPMORBIT_H__

/* -----------------------------------------------------------------------------
// includes
// -------------------------------------------------------------------------- */
#include <math.h>
#include <bpm/bpm_defs.h>
#include <bpm/bpm_interface.h>

/* -----------------------------------------------------------------------------
// macro definitions
// -------------------------------------------------------------------------- */

/* -----------------------------------------------------------------------------
// typedefs, enums and other declarations
// -------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif

  /**
     Structure representing a 3-vector, for use in the orbit 
     generation routines
  */
  struct v3 { 
    double x; /**< x-coordinate */
    double y; /**< y-coordinate */
    double z; /**< z-coordinate */
  };
  

  /**
     Structure representing a 3x3-matrix, for use in the orbit 
     generation routines
  */
  struct m33 {
    double e[3][3]; /**< the matrix */
  };



/* -----------------------------------------------------------------------------
// function prototypes and declarations
// -------------------------------------------------------------------------- */


   /** 
      Generate the beam at the bpm position, so takes the coordinates from the 
      bpm in the global from (stored in bpm->geom_pos and bpm->geom_tilt and
      fills the local hit positions for the beam in the bpm, beam->bpmhit...
      Also transports the energy, charge etc.. through to this point...
      
      - generates the beam at bpm position
      - transports the energy, charge
      - sets the bunch arrival time in each cavity, offsetted by digi_trigtimeoffset 
        in the bpmconf

      @param beam the beam configuration
      @param bpm  the bpm confguration 
      @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */                 
  EXTERN int generate_bpm_orbit( beamconf_t* beam, bpmconf_t* bpm );
  


  /**
    Fill the beamconf structures with the lab coords of a corrector scan 

    @param bpm A bpmconf structure containing the info about the BPM
    @param beam The beamconf structure that contains the beam info  
    @param num_evts The number of events in each corrector scan step 
    @param num_steps The number of corrector scan steps  
    @param angle_range The angle over which the corrector scan is performed (in urad)  
    @param angle The orientation (from the horizontal) of the corrector scan axis (in urad) 
    @param z_pos The z position in the beamline of the corrector 
    @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */   
  EXTERN int generate_corr_scan( bpmconf_t* bpm, beamconf_t* beam, int num_evts,
				 int num_steps, double angle_range, double angle,
				 double z_pos );

  
  /** 
    Fill the beamconf structures with the lab coords of a mover scan. At present,
    this just changes the beam coords rather than the physical bpm coords. In the
    future, should add the possiblility of time-varying BPM positions
     
    @param beam The beamconf structure that contains the beam info
    @param num_evts The number of events in each corrector scan step
    @param num_steps The number of corrector scan steps
    @param mover_range The size of the move (in um)
    @param angle The orientation (from the horizontal) of the mover scan axis (in urad)
    @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */
  EXTERN int generate_mover_scan( beamconf_t* beam, int num_evts, int num_steps, 
				  double mover_range, double angle );

  

  /** Copy 3-vector v2 into 3-vector v1 */
  void   v_copy(struct v3 *v1, struct v3 *v2);

  /** Return the magnitude of 3-vector v1 */
  double v_mag(struct v3 *v1);

  /** Scale 3-vector v1 with factor dscale */
  void   v_scale(struct v3 *v1, double dscale);

  /** Normalise 3-vector v1 to unit vector */
  void   v_norm(struct v3 *v1);

  /** Multiply matrix m1 with 3-vector v1 : m1.v1, result is in v1 */
  void   v_matmult(struct m33 *m1, struct v3 *v1);

  /** Add two 3-vectors v1 and v2, result is in v1 */
  void   v_add(struct v3 *v1, struct v3 *v2);

  /** Subtract 3-vectors v1 - v2, result is in v1*/
  void   v_sub(struct v3 *v1, struct v3 *v2);

  /** Return Scalar product of 3-vectors v1 and v2 */
  double v_dot(struct v3 *v1, struct v3 *v2);

  /** Return the vector product of 3 vectors v1 x v2, result is in v1 */
  void   v_cross(struct v3 *v1, struct v3 *v2);

  /** Print the 3-vector to stdout */
  void   v_print(struct v3 *v1);

  /** Create rotation 3x3 matrix with the 3 euler angles alpha, beta and gamma, result in m1 */
  void   m_rotmat(struct m33 *m1, double alpha, double beta, double gamma);

  /** 3x3 Matrix multiplication m1.m2, result in m*/
  void   m_matmult(struct m33 *m, struct m33 *m1, struct m33 *m2);

  /** 3x3 Matrix addition m1+m2, result in m1 */
  void   m_matadd(struct m33 *m1, struct m33 *m2);

  /** Print 3x3 matrix m1 to stdout */
  void   m_print(struct m33 *m1);
  
#ifdef __cplusplus
}
#endif

#endif /* #ifndef BPMORBIT_H__ */
/* @} */
/* ================================ end of file ============================= */
