/**
   @file
   @ingroup processing
*/

#include <math.h>
#include <bpm/bpm_messages.h>
#include <bpm/bpm_process.h>

/**
   Performs one step in the gaussian filter
   
   @param ddc  the complex ddc wavform
   @param ns   number of samples
   @param fs   sampling frequency
   @param istart starting sample for moving window
   @param istop  stop stample for moving window
   @param tfilter filter time
   @param filtBW  filter bandwith
   @param out    a double[2] that will contain the resulting filtered Re and Im values
                 at tfilter
   
   @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
*/
int ddc_gaussfilter_step( double **ddc, int ns, double fs, int istart, int istop, 
			  double tfilter, double filtBW, double *out ) {
  
  int i;
  double ti, omega;

  if ( ! out ) {
    bpm_error( "Invalid pointer in gaussfilter_step(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }
  
  if ( istart < 0 ) {
    istart = 0;
    bpm_warning( "Setting istart to 0 in gaussfilter_step(...)", __FILE__, __LINE__ );
  }

  if ( istop > ns ) {
    istop = ns;
    bpm_warning( "Setting istop to ns in gaussfilter_step(...)", __FILE__, __LINE__ );
  }
  
  // reset values
  out[Re] = 0.;
  out[Im] = 0.;

  // convert filter bandwith from MHz to rad/usec
  omega = 2. * PI * filtBW;

  // perform one filter step between istart and istop
  for( i=istart; i<istop; i++ ) {

    if ( sample_to_time( fs, ns, i, &ti ) == BPM_FAILURE ) {
      bpm_error( "Unable to get time from sample in gaussfilter_step(...)",
		 __FILE__, __LINE__ );
      return BPM_FAILURE;
    }

    out[Re] += ddc[i][Re] * omega * exp( -SQR( omega * ( ti - tfilter) ) / 2. ) / sqrt( 2.*PI );
    out[Im] += ddc[i][Im] * omega * exp( -SQR( omega * ( ti - tfilter) ) / 2. ) / sqrt( 2.*PI );

  }

  return BPM_SUCCESS;
}
