/**
   @file
   @ingroup processing
*/
#include <stdio.h>
#include <stdlib.h>

#include <bpm/bpm_alloc.h>
#include <bpm/bpm_units.h>
#include <bpm/bpm_messages.h>
#include <bpm/bpm_process.h>

#ifndef DDC_FULL_DEBUG__
#define DDC_FULL_DEBUG__
#endif

#undef DDC_FULL_DEBUG__



/**
   Does the DDC of the full waveform and stores it into the ampwf and phasewf waveforms
   this routine calls the simple ddc(...) routine to do one step. Note that this one
   doesn't need t0 or t0Offset as it will scan through the entire waveform...
   
   @param wf  the waveform
   @param ns  the number of samples
   @param nbits the number of digitiser bits
   @param fs the sampling frequency
   @param t0 the trigger time
   @param freq the frequency of the waveform to downmix with
   @param tdecay the decay time of the waveform
   @param filtBW the gaussian filter bandwith
   @param epsFilt the gaussian filter cut-off parameter
   @param out contains the downconverted, filtered complex waveform
   
   @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
*/
int ddc_waveform( int *wf, int ns, int nbits, double fs, double t0,
		  double freq, double tdecay, double filtBW, double epsFilt,
		  double **out ) {

  double **ddc;
  double *wfmod;
  int i;
  double ped, rms;

  if ( ! wf ) {
    bpm_error( "Invalid waveform pointer in ddc_waveform(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }
  
  // allocate double array to store the wf
  wfmod = alloc_simple_wave_double( ns );
  if ( ! wfmod ) {
    bpm_error( "Unable to allocate memory for wfmod in ddc_waveform(...)",
               __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // allocate complex ddc waveform
  ddc = alloc_complex_wave_double( ns );
  if ( ! ddc ) {
    bpm_error( "Unable to allocate memory for ddc in ddc_waveform(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // get pedestal
  if ( get_pedestal( wf, ns, 20, &ped, &rms ) == BPM_FAILURE ) {
    bpm_error( "Unable to retreive the pedestal in downmix_waveform(...)",
             __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // remove pedestal

  for (i=0;i<ns;i++) wfmod[i] = wf[i] - ped;

  // downmix the waveform
  if ( downmix_waveform( wfmod, ns, fs, freq, t0, ddc ) == BPM_FAILURE ) {
    bpm_error( "Error in downmixing in ddc_sample_waveform(...)", 
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  if ( ddc_gaussfilter( ddc, ns, fs, filtBW, epsFilt, out ) == BPM_FAILURE ) {
    bpm_error( "Unable to filter the waveform in ddc_waveform(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // free up
  free_simple_wave_double( wfmod );
  free_complex_wave_double( ddc, ns );

  return BPM_SUCCESS;
}
