/**
   @file
   @ingroup processing
*/
#include <stdio.h>
#include <bpm/bpm_messages.h>
#include <bpm/bpm_process.h>

/**
   Process dipole waveform
   
   @param bpm Configuration structure for the bpm waveform to be processed
   @param cal Calibration strcture with calib info to use
   @param sig The BPM signal itself
   @param proc The resuling processed signal
   @param trig The already processed trigger waveform
   @param ref  The already processed reference waveform
   @param mode Processing mode

   @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
*/
int process_dipole( bpmconf_t *bpm, bpmcalib_t *cal, bpmsignal_t *sig, bpmproc_t *proc,
		    bpmproc_t *trig, bpmproc_t *ref, unsigned int mode ) {

  double I,Q;
  char msg[128];

  if ( ! bpm || ! sig || ! proc || ! trig || ! ref ) {
    bpm_error( "Invalid pointer arguments in process_dipole(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }
    
  if ( process_waveform( dipole, bpm, cal, sig, proc, trig, mode ) == BPM_FAILURE ) {
    sprintf( msg, "Unable to process waveform for BPM %s in process_dipole(...)", bpm->name );
    bpm_error( msg, __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // the structures should be filled now with amplitude and phases, let's compute
  // I,Q and the postion and tilts... for fit and ddc
  if ( proc->ddc_success && ref->ddc_success ) {

    I = Q = 0.;

    // process info for DDC
    if ( get_IQ( proc->ddc_amp, proc->ddc_phase, ref->ddc_amp, ref->ddc_phase, 
		 &Q, &I ) == BPM_FAILURE ) {
      sprintf( msg, "Could not calculated I and Q for BPM %s in process_dipole(...)", bpm->name );
      bpm_error( msg, __FILE__, __LINE__ );
      return BPM_FAILURE;
    } 

    proc->ddc_I = I;
    proc->ddc_Q = Q;

    if ( get_pos( Q, I, cal->IQphase, cal->posscale, &(proc->ddc_pos) ) == BPM_FAILURE ) {
      sprintf( msg, "Could not get ddc position for BPM %s in process_dipole(...)", bpm->name );
      bpm_error( msg, __FILE__, __LINE__ );
      return BPM_FAILURE;
    }

    if ( get_slope( Q, I, cal->IQphase, cal->slopescale, &(proc->ddc_slope) ) == BPM_FAILURE ) {
      sprintf( msg, "Could not get ddc slope for BPM %s in process_dipole(...)", bpm->name );
      bpm_error( msg, __FILE__, __LINE__ );
      return BPM_FAILURE;
    }

  } /* if ( proc->ddc_success ) */
  
  // repeat the excersice for the fit
  if( proc->fit_success && ref->fit_success ) {

    I = Q = 0.;

    // process info for fit
    if ( get_IQ( proc->fit_amp, proc->fit_phase, ref->fit_amp, ref->fit_phase, 
		 &Q, &I ) == BPM_FAILURE ) {
      sprintf( msg, "Could not calculated I and Q for BPM %s in process_dipole(...)", bpm->name );
      bpm_error( msg, __FILE__, __LINE__ );
      return BPM_FAILURE;
    } 

    proc->fit_I = I;
    proc->fit_Q = Q;

    if ( get_pos( Q, I, cal->IQphase, cal->posscale, &(proc->fit_pos) ) == BPM_FAILURE ) {
      sprintf( msg, "Could not get fit position for BPM %s in process_dipole(...)", bpm->name );
      bpm_error( msg, __FILE__, __LINE__ );
      return BPM_FAILURE;
    }
    
    if ( get_slope( Q, I, cal->IQphase, cal->slopescale, &(proc->fit_slope) ) == BPM_FAILURE ) {
      sprintf( msg, "Could not get fitted slope for BPM %s in process_dipole(...)", bpm->name );
      bpm_error( msg, __FILE__, __LINE__ );
      return BPM_FAILURE;
    }
    
  } /* if( proc->fit_success ) */


  return BPM_SUCCESS;
}

// end of file
