/**
   @file
   @ingroup sim
   @brief   libbpm waveform simulation routines

   This header contains the definitions for the libbpm RF waveform
   simulation routines
*/

/** @addtogroup sim */
/** @{ */

#ifndef BPMSIM_H__
#define BPMSIM_H__

/* -----------------------------------------------------------------------------
// includes
// -------------------------------------------------------------------------- */
#include <math.h>
#include <bpm/bpm_defs.h>
#include <bpm/bpm_interface.h>

/* -----------------------------------------------------------------------------
// macro definitions
// -------------------------------------------------------------------------- */

/* -----------------------------------------------------------------------------
// typedefs, enums and other declarations
// -------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
// function prototypes and declarations
// -------------------------------------------------------------------------- */

  EXTERN int generate_monopole( bpmconf_t*, beamconf_t*, bpmsignal_t* );

  EXTERN int generate_dipole( bpmconf_t*, beamconf_t*, bpmsignal_t* );

  EXTERN int generate_diode( bpmconf_t*, beamconf_t*, bpmsignal_t* );
  

  EXTERN int get_monopole_response( double bunchcharge, double chargesens,
				    double arrivaltime, double cavityfreq,
				    double *amp, double *phase );

  EXTERN int get_dipole_response( double bunchcharge, double chargesens, 
				  double pos, double possens, 
				  double tilt, double tiltsens,
				  double arrivaltime, double cavityfreq,
				  double *amp, double *phase );
  /*
    so rougly same, but... pos/tilt rotation.. extra phase pickup ????,.... 
  */

  EXTERN int get_dipole_amp( double bunchcharge, double bunchlength,
                             double pos, double possens,
                             double slope, double slopesens,
                             double tilt, double tiltsens,
                             double *amp, double *phase );
  /*
    generates the complex amplitude of the dipole mode
  */

  EXTERN int get_monopole_amp( double bunchcharge, double bunchlength,
                             double chargesens,
                             double *amp, double *phase );
  /*
    generates the complex amplitude of the monopole mode
  */

  EXTERN int add_excitation( double ttrig, double *RF );

  /**
     @ingroup simulation

     Generates a step to excite the resonator
  */

  EXTERN int simple_wave( double amp, double phase, double ttrig, double freq, double tdecay,
			  double ped, double ampnoise, double phasenoise,
			  double fs, int nbits, int *wf, int ns );
  EXTERN int simple_tone( double amp, double phase, double freq,
			  double ped, double ampnoise, double phasenoise,
			  double fs, int nbits, int *wf, int ns );
  /*
    makes a simple digitised waveform and tone without much RF hassle, just quick and dirty
  */

  EXTERN int add_wave( double amp, double phase, double freq, double ttrig, double tdecay,
		       double **RF );
  /*
    adds a wave to the complex wf array having the requred parameters, the wf representation
    will have rf_nsamples and rf_samplefreq as sampling freq internal to the software,
  */

  EXTERN int add_waveforms( double *RF, double *RFadd, double factor );
  /*
    adds complex waveform RFadd to complex waveform RF
  */

 
  EXTERN int reset_complex_wave( double **RF );
  /*
    resets a complex array
  */

  EXTERN int reset_simple_wave( int ns, double *wf );
  /*
    resets a double array
  */

  EXTERN int get_real_part( double **RF, double *wf );
  /*
    returns an array in wf containing the real part of RF
  */

  EXTERN int get_imaginary_part( double **RF, double *wf );
  /*
    returns an array in wf containing the imaginary part of RF
  */
 
  EXTERN int get_amplitude( double **RF, double *wf );
  /*
    returns an array in wf containing the amplitudes of RF
  */

  EXTERN int get_phase( double **RF, double *wf );
  /*
    returns an array in wf containing the phase of RF
  */

  EXTERN int get_complex_from_ReIm( double *RF_Re, double *RF_Im, double **RF );
  /*
    returns a complex array combined from the real and imaginary parts
  */
  EXTERN int get_complex_from_AmpPhi( double *Amp, double *Phi, double **RF );
  /*
    returns a complex array restored from the amplitude and phase
  */


  EXTERN int add_amplnoise( double amplnoise, double *IF_Re, double *IF_Im );
  /*
    add ampl noise to the IF waveform - noise given in volts
  */


  EXTERN int digitise( double *IF, int nbits, double fs, double range_min,
	               double range_max, int ns, int *wf );
  /*
    digitises the IF with the given nbits, sampling freq and number of samples, 
    returns *wf, takes care of saturation (as we have the nbits...)
  */


  /*
    generate diode just calls rf_rectify or so to generate a pulse
  */
  

#ifdef __cplusplus
}
#endif

#endif /* #ifndef BPMINTERFACE_H__ */
/** @} */
/* ================================ end of file ============================= */
