/**
   @file
   @ingroup sim
*/
#include <bpm/bpm_messages.h>
#include <bpm/bpm_simulation.h>
#include <bpm/bpm_rf.h>
#include <bpm/bpm_nr.h>

/**
   Digitises the waveform 
   
   @param IF The input waveofrm to digitse
   @param nbits The number of bits of the ADC
   @param fs the sampling frequency
   @param range_min the minimum voltage and
   @param range_max the maximum voltage the ADC can process
   @param ns number of samples in the sampled waveform
   @param wf sampled waveform
   
   @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
*/
int digitise( double *IF, int nbits, double fs, double range_min,
	      double range_max, int ns, int *wf ) {

  if (( ! IF ) || ( ! wf )) {
    bpm_error( "Invalid pointer arguments in digitise(...)", 
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }
  
  if ( nbits <= 0 ) {
    bpm_error( "Invalid number of ADC bits to digitise(...)",
	       __FILE__, __LINE__);
    return BPM_FAILURE;
  }

  double range, t_adc, amp, t_rf;
  int i,j;

  range = pow( 2., nbits ); // range of the ADC
  
  // find trigger time in the waveform to get things started
  i = 0;
  t_rf = 0.;

  // now at first sample
  for ( j = 0; j < ns; j++ ) {

    // ADC timestamp of this sample, (without time jitter)
    t_adc = (double) j / fs;
    
    // add time jitter here when required
    //t_adc = nr_rangauss( t_adc, jit );

    // find the two samples inbetween which this ADC sampling time is located
    while( t_rf < t_adc ) {
      t_rf = i / rf_samplefreq;
      i++;
    }

    // Interpolate where necessary
    if ( i >= rf_nsamples ) { 

      // in this case we just fill the array with the pedestal 
      // but to overcome weirdness in say the FFT, we add 0.5 count RMS noise 
      wf[j] = (int) range/2 + nr_rangauss( 0., .5 );

    } else {

      if ( i > 0 ) {

        // interpolate IF[i][Re] linearily between these two values at t_adc
        // so here we get the signal amplitude
        amp = IF[i-1] + ( IF[i] - IF[i-1] ) * 
	  ( t_adc - (t_rf - 1. / rf_samplefreq ) ) * rf_samplefreq;

      } else {

        amp = IF[i];

      }

      // now convert this value to an ADC value, adding half of the range
      // as pedestal
      wf[j] = (int) ( amp * range/(range_max-range_min) + range/2) + nr_rangauss( 0., .5 );

      // now address saturation in the ADC 
      if( wf[j] < 0 ) wf[j] = 0;
      if( wf[j] > ( range - 1 ) ) wf[j] = range - 1;

    }
  }

  return BPM_SUCCESS;
}

// end of file
