/**
   @file
   @ingroup sim
*/
#include <bpm/bpm_messages.h>
#include <bpm/bpm_simulation.h>
#include <bpm/bpm_rf.h>
#include <bpm/bpm_alloc.h>


/**
   Generate a diode waveform using the given bpm parameters. Essentially, a 
   reference waveform is generated and then rectified to produce the 
   trigger pulse.
   
   @param bpm Structure containing the BPM info
   @param beam Structure containing the beam hit info
   @param sig Structure for where to place the digitised waveform
   
*/
int generate_diode( bpmconf_t *bpm, beamconf_t *beam, bpmsignal_t *sig ) {


  if ( ! bpm || ! beam || ! sig ) {
    bpm_error( "Invalid pointer arguments in generate_diode(...)", 
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // Compute the monopole response of the BPM
  double amp, phase;
  get_monopole_response( beam->charge, bpm->cav_chargesens, beam->arrival_time, bpm->cav_freq,
			 &amp, &phase );
  int i;
  // Generate the waveoform
  double **RF = alloc_complex_wave_double( rf_nsamples );
  add_wave( amp, phase, bpm->cav_freq, beam->arrival_time + bpm->digi_trigtimeoffset, bpm->cav_decaytime, RF );

  // Rectify the waveform
  rf_rectify( RF );

  // Filter the signal
  rf_filter( RF, bpm->rf_filtertype, bpm->rf_nfiltpars, bpm->rf_filterpars );
  
  // Amplify
  //rf_amplify( RF, bpm->rf_gain );

  // Add some noise - currently assume ADC goes from -1V -> +1V
  double conv = (double) pow( 2, bpm->digi_nbits ) / 2.;
  double noise = bpm->digi_ampnoise / conv;
  //add_amplnoise( noise, RF );

  // digitise the waveform - currently assume ADC goes from -1V -> +1V
  int *wf = alloc_simple_wave_int( bpm->digi_nsamples );
  //digitise( RF, bpm->digi_nbits, bpm->digi_voltageoffset, bpm->digi_freq, conv, 
//	    bpm->digi_nsamples, wf);

  // Finally, store the waveform in the signal structure
  sig->wf = wf;
  sig->ns = bpm->digi_nsamples;

  // Clean up
  free_complex_wave_double( RF, rf_nsamples );

  return BPM_SUCCESS;
}

// end of file
