/**
   @file
   @ingroup sim
*/
#include <math.h>

#include <bpm/bpm_messages.h>
#include <bpm/bpm_nr.h>
#include <bpm/bpm_units.h>
#include <bpm/bpm_simulation.h>

/**
   Just fills an array of integers straight from the parameters given, this
   routine is a little tool to quickly generate digitised decaying waveforms
   without much RF hassle. The routine overwrites the wf array, so doesn't add
   the values...
   
   The waveform shape is
   @f[ A e^{-(t-t0)/\tau} sin( 2\pi\nu(t-t0) + \phi ) + offset @f]
   with added phasenoise to the phase \phi and amplitude noise to the offset.
   The user needs to give the nbits in the ADC as well to enable the routine
   to cut off the sample and have saturation...
   
   @param amp      amplitude in ADC channels
   @param phase    phase of waveform
   @param ttrig    trigger time of waveform (t0)
   @param freq     frequency of waveform
   @param tdecay   decay time of waveform
   @param ped        offset of waveform in ADC channels
   @param ampnoise   amplitude noise in ADC channels
   @param phasenoise phase noise 
   @param fs         sampling frequency
   @param nbits      number of bits in the ADC
   @param *wf        the returned waveform
   @param  ns        the number of samples to fill
   
   @return BPM_SUCCESS upon success, BPM_ERROR upon error
*/
int simple_wave( double amp, double phase, double ttrig, double freq, 
		 double tdecay, double ped, double ampnoise, double phasenoise,
		 double fs, int nbits, int *wf, int ns ) {
  // starttime for sampling, let's have it at 0 for the moment,
  // maybe later on we need to put it in the routine as an argument
  double t0S = 0. * usec; 
  double w, t, max;
  int i;

  if ( ! wf ) {
    bpm_error( "Invalid pointer argument in simple_wave(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }
  
  // maximum sample
  max = pow( 2., (double) nbits );

  for ( i=0; i<ns; i++ ) {
    // calculate sampling time from the sample number and the 
    // sampling frequency, we start sampling at t=t0S
    t = t0S + ( (double) i + 0.5 ) / fs;
    
    w = 0.;
    if ( t >= ttrig ) {
      w = amp 
	* exp( - ( t - ttrig ) / tdecay ) 
	* sin( 2. * PI * freq * ( t - ttrig ) + nr_rangauss( phase, phasenoise ) );
    } 

    // add the voltage offset + noise
    w += nr_rangauss( ped, ampnoise );

    // will the waveform
    if ( w < 0. ) wf[i] = 0;
    else if ( w > max ) wf[i] = (int) max;
    else wf[i] = (int) w;

  } /* for ( i=0; i<ns; i++ ) */
  
  return BPM_SUCCESS;
}

// end of file
