/**
   wfprocess

   libbpm example program that takes waveforms and processes them with routines based
   upon libbpm

   Author : Bino Maiheu, University College London (2006-2007)
            Mark Slater, University of Cambridge (2005-2007)

   ChangeLog :
*/

#include <stdio.h>
#include <stdlib.h>

#include <bpm/bpm_units.h>
#include <bpm/bpm_interface.h>
#include <bpm/bpm_process.h>
#include <bpm/bpm_orbit.h>
#include <bpm/bpm_nr.h>
#include <bpm/bpm_calibration.h>
#include <bpm/bpm_analysis.h>

#define GENERATE

int main( int argc, char **argv ) {

  /*---------------------------------------------------------------------------------------
    INITIALISE

    Set up the bpm and beam configuration files to be used. Based on the ESA beamline.
    ---------------------------------------------------------------------------------------
  */
  
  int num_bpms;
  bpmconf_t *beamline;

  // Load in the esa configuration file
  load_bpmconf( "examples/data/esa_bpm_conf.asc", &beamline, &num_bpms );

  /*---------------------------------------------------------------------------------------
    CREATE THE BEAM HITS

    Produce two corrector scans, a mover scan and some resolution data. These routines fill
    the lab coords of a set of beamconf_t structures. These are then transferred to local
    BPM coords.

    NOTE: This would normally be created by a full GEANT simulation. Here, we're just using the
    basic routines
    ---------------------------------------------------------------------------------------
  */
  
  const int num_events = 400;                 // The number of events in the whole run
  beamconf_t bpmhits[num_bpms][num_events];    // The structures that will contain the hits for each bpm
  int ibpm, evt;
  double energy, charge, pos_x, pos_y, tilt_x, tilt_y, z_pos;
  bpmsignal_t wfs[num_bpms][num_events];
  FILE *outfile;
  int cal_start = 0, cal_steps = 5, cal_num = 10;

#ifdef GENERATE
  
  // Set up the energy and charge of the beam
  for ( evt = 0; evt < num_events; evt++) {
    
    energy = nr_rangauss( 10. * GeV, 0.2 * GeV );
    charge = nr_rangauss( 1.6 * nC, 0.2 * nC);

    // Jitter parameters
    pos_x = nr_rangauss( 0., 2. );
    pos_y = nr_rangauss( 0., 2. );
    tilt_x = nr_rangauss( 0., 0.05 * urad );          // x'/dy shift (about y axis)
    tilt_y = nr_rangauss( 0., 0.05 * urad );          // y'/dx shift (about x axis)
    z_pos = 0.;                                     // beam waist position

    for ( ibpm = 0; ibpm < num_bpms; ibpm++ ) {
      
      bpmhits[ibpm][evt].energy = energy;
      bpmhits[ibpm][evt].sig_energy = 0.1 * GeV;
      bpmhits[ibpm][evt].charge = charge;
      bpmhits[ibpm][evt].sig_charge = 0.1e-9;

      // Generate some beam jitter
      bpmhits[ibpm][evt].beampos[0] = pos_x + tilt_x * ( beamline[ibpm].geom_pos[2] - z_pos );
      bpmhits[ibpm][evt].beampos[1] = pos_y + tilt_y * ( beamline[ibpm].geom_pos[2] - z_pos );

      // Calculate arrival time in microsecs
      bpmhits[ibpm][evt].arrival_time = ( beamline[ibpm].geom_pos[2] / 3.e8 ) * 1.e6;
    }
  }  

  // Now generate the beam positions
  for (ibpm = 0; ibpm < num_bpms; ibpm++) {

    // Corrector Scans
    generate_corr_scan( &(beamline[ibpm]), &(bpmhits[ibpm][cal_start]), 
			cal_num,        // Number of events per step
			cal_steps,          // number of steps
			4 * urad,   // angle range over which to scan
			0 * urad,   // angle of scan (this is an X scan)
			-45.        // z position of corrector 
			);

    generate_corr_scan( &(beamline[ibpm]), &(bpmhits[ibpm][cal_start + cal_num * cal_steps]), 
			cal_num,            // Number of events per step
			cal_steps,              // number of steps
			4 * urad,       // angle range over which to scan
			PI / 2. * rad, // angle of scan (this is an Y scan)
			-45.            // z position of corrector 
			);
  }
  
  // Mover Scans
  generate_mover_scan( &(bpmhits[0][cal_start + cal_num * cal_steps * 2]), 
		       cal_num,       // Number of events per step
		       cal_steps,         // number of steps
		       500,       // range over which to scan (in um)
		       0 * urad   // angle of scan (this is an X scan)
		       );
  
  generate_mover_scan( &(bpmhits[0][cal_start + cal_num * cal_steps * 3]), 
		       cal_num,            // Number of events in scan
		       cal_steps,              // number of steps
		       500,            // range over which to scan (in um)
		       PI / 2. * rad  // angle of scan (this is an Y scan)
		       );

  // Finally, transform from lab coords to BPM local frame. This step would need to be done
  // after GEANT had filled the lab coords of beam hits
  for (ibpm = 0; ibpm < num_bpms; ibpm++) {
    for ( evt = 0; evt < num_events; evt++ ) {
      
      generate_bpm_orbit( &(bpmhits[ibpm][evt]), &(beamline[ibpm]) );

    }
  }

  // Output the positions as recorded at BPM 3
  outfile = fopen("examples/data/BPMHits.asc", "w");
  for ( evt = 0; evt < num_events; evt++ ) {
      fprintf( outfile, "%d %f\n", evt, bpmhits[0][evt].bpmhit[0] );
    }
  fclose(outfile);

  /*---------------------------------------------------------------------------------------
    GENERATE THE WAVEFORMS
    
    Loop over all the events and create all the digitised waveforms from the run created above.
    ---------------------------------------------------------------------------------------
  */
  printf("Generating waveforms...\n");

  for ( evt = 0; evt < num_events; evt++ ) {//num_events; evt++) {
 
    for ( ibpm = 0; ibpm < num_bpms; ibpm++) {
      
      switch( beamline[ibpm].cav_type ) {
	
      case diode:
	generate_diode( &(beamline[ibpm]), &(bpmhits[ibpm][evt]), &(wfs[ibpm][evt]) );
	break;

      case monopole:
	generate_monopole( &(beamline[ibpm]), &(bpmhits[ibpm][evt]), &(wfs[ibpm][evt]) );
	break;

      case dipole:
	generate_dipole( &(beamline[ibpm]), &(bpmhits[ibpm][evt]), &(wfs[ibpm][evt]) );
	break;
      }
    }

    if ((evt % 10) == 0) printf("Event %d\n", evt);
  }

  printf("Done\n");

  // Save the waveforms
  for (ibpm = 0; ibpm < num_bpms; ibpm++ ) {

    char buf[255];
    sprintf(buf, "examples/data/BPM%dWFs.asc", ibpm);
    save_signals( buf, wfs[ibpm], num_events );
    
  }

#endif

  /*---------------------------------------------------------------------------------------
    PROCESS THE WAVEFORMS - FREQUENCY FITTING
    
    First, we need to determine the frequencies and decay constants of the generated
    waveforms.
    
    NOTE: This would be where to start with real data, after having filled the bpmsignal_t's
    ---------------------------------------------------------------------------------------
  */

  printf("Fitting for frequencies and decay constants...\n");

  bpmsignal_t *wfs2[num_bpms];

  // Load the waveforms saved above
  for (ibpm = 0; ibpm < num_bpms; ibpm++ ) {

    char buf[255];
    sprintf(buf, "examples/data/BPM%dWFs.asc", ibpm);
    load_signals( buf, &(wfs2[ibpm]) );
  }
  
  // Output example waveforms
  int samp, pr_evt = 189;
  outfile = fopen("examples/data/DiodeWF.asc", "w");
  for ( samp = 0; samp < wfs2[2][pr_evt].ns; samp++ ) {
    fprintf( outfile, "%d %i\n", samp, wfs2[2][pr_evt].wf[samp] );
  }
  fclose(outfile);
  
  outfile = fopen("examples/data/MonopoleWF.asc", "w");
  for ( samp = 0; samp < wfs2[3][pr_evt].ns; samp++ ) {
    fprintf( outfile, "%d %i\n", samp, wfs2[3][pr_evt].wf[samp] );
  }
  fclose(outfile);
  
  outfile = fopen("examples/data/DipoleWF.asc", "w");
  for ( samp = 0; samp < wfs2[0][pr_evt].ns; samp++ ) {
    fprintf( outfile, "%d %i\n", samp, wfs2[0][pr_evt].wf[samp] );
  }
  fclose(outfile);

  bpmproc_t proc_data[num_bpms][num_events];
  bpmcalib_t cal_data[num_bpms];

  // Set up the initial calibrations
  for ( ibpm = 0; ibpm < num_bpms; ibpm++) {

    cal_data[ibpm].t0Offset = 0.2 * usec;
    cal_data[ibpm].ddcfiltBW = 4.* MHz;
    cal_data[ibpm].ddcepsFilt = 0.001;
  }

  for ( evt = 0; evt < num_events; evt++ )
    {      
      // Process all bpms in this order: diode, reference, dipole
      for ( ibpm = 0; ibpm < num_bpms; ibpm++ ) {
	if ( beamline[ibpm].cav_type == diode ) {
	  process_diode( &(beamline[ibpm]), &(wfs2[ibpm][evt]), &proc_data[ibpm][evt]);	
	}

      }

      // Now refs
     for ( ibpm = 0; ibpm < num_bpms; ibpm++ ) {
	if ( beamline[ibpm].cav_type == monopole ) {
	  process_monopole( &(beamline[ibpm]), &(cal_data[ibpm]), &(wfs2[ibpm][evt]), 
			    &(proc_data[ibpm][evt]), 
			    &(proc_data[beamline[ibpm].diode_idx][evt]), // Trigger data
			    PROC_DO_FIT );
	}
     }

     // Finally dipoles
     for ( ibpm = 0; ibpm < num_bpms; ibpm++ ) {
	if ( beamline[ibpm].cav_type == dipole ) {
	  process_dipole( &(beamline[ibpm]), &(cal_data[ibpm]), &(wfs2[ibpm][evt]), 
			    &(proc_data[ibpm][evt]), 
			    &(proc_data[beamline[ibpm].diode_idx][evt]), // Trigger data
			    &(proc_data[beamline[ibpm].ref_idx][evt]),   // Reference data
			    PROC_DO_FIT );
	}       
     }
    }

  // Calculate the mean frequencies and decay constants
  printf("Frequencies and Decay Constants for the BPMS:\n");

  for (ibpm = 0; ibpm < num_bpms; ibpm++ ) {

    if ( beamline[ibpm].cav_type != diode ) {
      double freq[num_events], tdecay[num_events];
      
      for ( evt = 0; evt < 10; evt++ ) {
	// Transfer to temporary storage
	freq[evt] = proc_data[ibpm][evt].fit_freq;
	tdecay[evt] = proc_data[ibpm][evt].fit_tdecay;
      }
      
      cal_data[ibpm].freq = nr_median( 10, freq );
      cal_data[ibpm].tdecay = nr_median( 10, tdecay );
      
      printf("BPM %d:  Freq: %f   TDecay: %f\n", ibpm, cal_data[ibpm].freq, cal_data[ibpm].tdecay);
    }
  }

  printf("Done.\n");

  /*---------------------------------------------------------------------------------------
    PROCESS THE WAVEFORMS - I's and Q's
    
    Loop over all the events, process the waveforms and find the I's and Q's to allow the
    calibration to be determined.

    ---------------------------------------------------------------------------------------
  */

  printf("Processing the waveforms...\n");

  for ( evt = 0; evt < num_events; evt++ )
    {
      // refs
      for ( ibpm = 0; ibpm < num_bpms; ibpm++ ) {
	if ( beamline[ibpm].cav_type == monopole ) {
	  process_monopole( &(beamline[ibpm]), &(cal_data[ibpm]), &(wfs2[ibpm][evt]), 
			    &(proc_data[ibpm][evt]), 
			    &(proc_data[beamline[ibpm].diode_idx][evt]), // Trigger data
			    PROC_DO_DDC | PROC_DO_FIT );
	}
      }
      
      // dipoles
      for ( ibpm = 0; ibpm < num_bpms; ibpm++ ) {
	if ( beamline[ibpm].cav_type == dipole ) {
	  process_dipole( &(beamline[ibpm]), &(cal_data[ibpm]), &(wfs2[ibpm][evt]), 
			  &(proc_data[ibpm][evt]), 
			  &(proc_data[beamline[ibpm].diode_idx][evt]), // Trigger data
			  &(proc_data[beamline[ibpm].ref_idx][evt]),   // Reference data
			  PROC_DO_DDC | PROC_DO_FIT );
	}
      }
    }
  
  // Output the amplitude response
  outfile = fopen("examples/data/AmpResponse.asc", "w");
  for ( evt = 0; evt < num_events; evt++ ) {
      fprintf( outfile, "%d %f\n", evt, proc_data[0][evt].ddc_amp );
    }
  fclose(outfile);

  // Output the Is and Qs for the run
  outfile = fopen("examples/data/IsAndQs.asc", "w");
  for ( evt = 0; evt < num_events; evt++ ) {
      fprintf( outfile, "%f %f\n", proc_data[0][evt].ddc_I, proc_data[0][evt].ddc_Q );
    }
  fclose(outfile);

  // Output the amplitude response
  outfile = fopen("examples/data/AmpResponseFit.asc", "w");
  for ( evt = 0; evt < num_events; evt++ ) {
    if ( (proc_data[3][evt].fit_amp > 0 ) && (proc_data[3][evt].fit_amp < 8000 ))
      fprintf( outfile, "%d %f\n", evt, proc_data[0][evt].fit_amp );
    }
  fclose(outfile);

  // Output the Is and Qs for the run
  outfile = fopen("examples/data/IsAndQsFit.asc", "w");
  for ( evt = 0; evt < num_events; evt++ ) {    
    if ( (proc_data[0][evt].fit_I > -100 ) && (proc_data[0][evt].fit_I < 100 ))
      fprintf( outfile, "%f %f\n", proc_data[0][evt].fit_I, proc_data[0][evt].fit_Q );
    }
  fclose(outfile);

  printf("Done.\n");

  /*---------------------------------------------------------------------------------------
    CALIBRATE THE BPMS

    Generate and save a calibration from the two corrector scans
    ---------------------------------------------------------------------------------------
  */
 
  printf("Calibrating...\n");

  beamconf_t calc_beampos[num_bpms][num_events];
  int startevt, stopevt;

  for (ibpm = 0; ibpm < num_bpms; ibpm++)
    {
      if ( beamline[ibpm].cav_type == dipole ) {

	if ( beamline[ibpm].cav_polarisation == horiz ) {

	  startevt = cal_start;
	  stopevt = cal_start + cal_num * cal_steps - 1;
	}
	  
	setup_calibration( &(beamline[ibpm]), &(proc_data[ibpm][startevt]), cal_num * cal_steps,
			   startevt, stopevt, 0. * urad, -170., 
			   170., cal_steps, &(calc_beampos[ibpm][startevt]) );
	
	calibrate( &(beamline[ibpm]), &(calc_beampos[ibpm][startevt]), &(proc_data[ibpm][startevt]), 
		   cal_num * cal_steps, &(cal_data[ibpm]) );
      }
    }

  save_calibration( "examples/data/ESACalib.asc", beamline, cal_data, num_bpms );

  printf("Done\n");

  /*---------------------------------------------------------------------------------------
    REPROCESS THE WAVEFORMS
    
    Loop over all the events and process the waveforms. At this point, no calibration is available
    
    NOTE: This would be where to start with real data, after having filled the bpmsignal_t's
    ---------------------------------------------------------------------------------------
  */

  printf("Reprocessing the Waveforms...\n");

  // Allow the tilts to be recorded
  for ( ibpm = 0; ibpm < num_bpms; ibpm++) {
    cal_data[ibpm].slopescale = 1.;
  }

  for ( evt = 0; evt < num_events; evt++ )
    {
  
      // t0 already found above
      
      // dipoles
      for ( ibpm = 0; ibpm < num_bpms; ibpm++ ) {
	if ( beamline[ibpm].cav_type == dipole ) {
	  process_dipole( &(beamline[ibpm]), &(cal_data[ibpm]), &(wfs2[ibpm][evt]), 
			  &(proc_data[ibpm][evt]), 
			  &(proc_data[beamline[ibpm].diode_idx][evt]), // Trigger data
			  &(proc_data[beamline[ibpm].ref_idx][evt]),   // Reference data
			  PROC_DO_DDC | PROC_DO_FIT );
	}
      }
    }

  // Output the positions
  outfile = fopen("examples/data/Positions.asc", "w");
  for ( evt = 0; evt < num_events; evt++ ) {
    fprintf( outfile, "%d %f\n", evt, proc_data[0][evt].ddc_pos );
    }
  fclose(outfile);

  printf("Done.\n");

  /*---------------------------------------------------------------------------------------
    CALCULATE THE RESOLUTION

    Now calculate the resoultion of the system
    ---------------------------------------------------------------------------------------
  */

  printf("Determining Resolution...\n");

  bpmproc_t *svd_confs[3];  // setup 3 bpms to compute resolution
  double coeffs[5];

  svd_confs[0] = &(proc_data[1][175]);
  svd_confs[1] = &(proc_data[0][175]);
  svd_confs[2] = &(proc_data[4][175]); 

  ana_get_svd_coeffs( svd_confs, 3, 200, 200, coeffs, ANA_SVD_TILT );

  double mean, rms;
  ana_compute_residual( svd_confs, 3, 200, coeffs, 
			ANA_SVD_TILT, &mean, &rms );
  
  printf("Final Resolution is:  %f\n", rms);

  return 0;
}
